// Copyright Joyent, Inc. and other Node contributors.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
// USE OR OTHER DEALINGS IN THE SOFTWARE.

'use strict';
const common = require('../common');
// Skip test in FreeBSD jails since 0.0.0.0 will resolve to default interface
if (common.inFreeBSDJail)
  common.skip('In a FreeBSD jail');

const assert = require('assert');
const dgram = require('dgram');

dgram.createSocket('udp4').bind(0, common.mustCall(function() {
  assert.strictEqual(typeof this.address().port, 'number');
  assert.ok(isFinite(this.address().port));
  assert.ok(this.address().port > 0);
  assert.strictEqual(this.address().address, '0.0.0.0');
  this.close();
}));

if (!common.hasIPv6) {
  common.printSkipMessage('udp6 part of test, because no IPv6 support');
  return;
}

dgram.createSocket('udp6').bind(0, common.mustCall(function() {
  assert.strictEqual(typeof this.address().port, 'number');
  assert.ok(isFinite(this.address().port));
  assert.ok(this.address().port > 0);
  let address = this.address().address;
  if (address === '::ffff:0.0.0.0')
    address = '::';
  assert.strictEqual(address, '::');
  this.close();
}));
